// Author: A.Voss@FH-Aachen.de
//
// idea
// tags: printf

#pragma GCC diagnostic ignored "-Wconversion-null"

#include <iostream>
#include <cstdlib>
#include <string>
#include <vector>
#include <stdexcept>
using namespace std;

#include "../../include/tools.h"

class C { };
class D : public C { };

class myexception 
{
    string _what;
public:
    myexception(const string & txt) : _what(txt) { }
    const char* what() const noexcept { return _what.c_str(); }
};

void f()
{
    try
    {
        throw 1;
    } 
    catch (...)
    {
        tools_log(); cout << "catch block in f" << endl;
        throw;
    }
}

class E 
{ 
public:
    E()  { tools_log(); cout << "in E::ctor" << endl; /* throw 1; */  }
    // do not throw in dtors!
    ~E() { tools_log(); cout << "in E::dtor" << endl; /* throw 1; */   }
    
    void g() throw(int)
    { 
        tools_log(); cout << "throw in g" << endl;
        // promise not to throw other than int, try 1.1
        throw 1;
    }
};

int main()
{
    // do some work, throw something in case of error
    tools_log(); cout << "throw 12.34" << endl;    
    try
    {
        throw 12.34;
        tools_log(); cout << "after throw" << endl;
    } 
    catch (...)
    {
        tools_log(); cout << "catch block" << endl;
    }

    // catch different types
    tools_log(); cout << "throw int" << endl;
    try
    {
        throw 23;
    } 
    catch (double d)
    {
        tools_log(); cout << "catch block double, d=" << d << endl;
    }
    catch (int n)
    {
        tools_log(); cout << "catch block int, n=" << n << endl;
    }
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }

    // throw objects - which catch becomes active?
    tools_log(); cout << "throw object" << endl;
    try
    {
        throw D(); 
    } 
    catch (const C &)
    {
        tools_log(); cout << "catch block C" << endl;
    }
    catch (const D &)
    {
        tools_log(); cout << "catch block D" << endl;
    }
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }

    // catch std exceptions
    tools_log(); cout << "div by zero? 1/0" << endl;
    try
    {
        double d1=1.0, d2=0.0, d3=d1/d2;
        //int n1=1, n2=0, n3=n1/n2;
        tools_log(); cout << "after throw (?) d3=" << d3 << endl;
        
        vector<double> v;
        // d3 = v[1];   // Speicherfehler! no exception!
        double d4 = v.at(1);
        tools_log(); cout << "after throw, d4=" << d4 << endl;
    } 
    catch (const out_of_range & e)
    {
        tools_log(); cout << "catch block out-of-range: '" << e.what() << "'" << endl;
    }
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }
    
    // catch user-defined exceptions
    tools_log(); cout << "throw myexception" << endl;
    try
    {
        throw myexception("oops");
    } 
    catch (const myexception & e)
    {
        tools_log(); cout << "catch block myexception: '" << e.what() << "'" << endl;
    }
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }

    // catch, rethrow, catch
    tools_log(); cout << "rethrow myexception" << endl;
    try
    {
        f();
    } 
    catch (int n)
    {
        tools_log(); cout << "catch block int, n=" << n << endl;
    }
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }

    // remember resources ...
    tools_log(); cout << "stack unwinding" << endl;
    try
    {
        E e;
        e.g();
    } 
    catch (...)
    {
        tools_log(); cout << "catch block rest" << endl;
    }
    
    return EXIT_SUCCESS;
}

